from typing import TypeVar
from .exceptions import *

ListOrDict = TypeVar('ListOrDict', list, dict)


class Stack:
    def __init__(self, inf=1, capacity=0):
        self.inf = inf
        if inf:
            self.__storage = []
        else:
            self.ind = 0
            self.__storage = [0 for _ in range(capacity)]
            self.capacity = capacity

    @property
    def storage(self):
        return self.__storage

    def push(self, elem):
        if self.inf:
            self.__storage.append(elem)
        else:
            if self.ind == self.capacity:
                raise StackError("Stack Is Full !")
            self.__storage[self.ind] = elem
            self.ind += 1

    def pop(self):
        if self.__storage:
            return self.__storage.pop()
        else:
            raise StackError("Stack Is Empty !")

    def display(self):
        return self.storage


class LIFO(Stack):
    pass


class Queue:
    def __init__(self, inf=1, capacity=0):
        self.inf = inf
        if inf:
            self.__storage = []
        else:
            self.ind = 0
            self.__storage = [0 for _ in range(capacity)]
            self.capacity = capacity

    @property
    def storage(self):
        return self.__storage

    def add(self, elem):
        if self.inf:
            self.__storage.append(elem)
        else:
            if self.ind == self.capacity:
                raise QueueError("Queue Is Full !")
            self.__storage[self.ind] = elem
            self.ind += 1

    def remove(self):
        if self.__storage:
            return self.__storage.pop(0)
        else:
            raise QueueError("Queue Is Empty !")

    def display(self):
        return self.storage


class FIFO(Queue):
    pass


class PyshaList(list):
    def __str__(self):
        s = "["
        for i in self:
            s += " " + str(i) + " ,"
        s = s[:-1] + "]"
        return s

    def __repr__(self):
        s = "["
        for i in self:
            s += " " + str(i) + " ,"
        s = s[:-1] + "]"
        return s

    def lshift(self, times):
        for _ in range(times):
            self = self[1:] + [self[0]]

    def rshift(self, times):
        for _ in range(times):
            self = [self[-1]] + self[:-1]

    # Just String,Int And Boolean Are Handled Here.
    def count_deep(self, tar):
        if not (type(tar) == int or type(tar) == str or type(tar) == bool):
            raise TypeError(
                "CountDeep Just Supports String,Integer And Boolean Types.")
        if not any(list(map(lambda x: type(x) == list or type(x) == set, self))):
            if type(tar) == str and any(list(map(lambda x: type(x) == str, self))):
                counted = 0
                for index in range(len(self)):
                    try:
                        counted += self[index].count(tar)
                    except:
                        continue
                return counted
            return self.count(tar)
        indexes = Stack()
        lists = Stack()
        target_list = self
        current_index = 0
        counted = 0
        while current_index != len(self):
            has_new_list = False
            for i in target_list:
                if type(tar) == type(i):
                    if tar == i:
                        counted += 1
                elif type(tar) != list:
                    indexes.push(current_index)
                    lists.push(target_list)
                    target_list = i.copy()
                    current_index = 0
                    has_new_list = True
                    break
                current_index += 1
            if not has_new_list:
                current_index = indexes.pop() + 1
                target_list = lists.pop()
        return counted

    def shift_index(self, index, left_shift=True, how_many=1):
        for _ in range(how_many):
            if not left_shift:
                if index == len(self) - 1:
                    self.rshift(1)
                else:
                    self[index], self[index +
                                      1] = self[index + 1], self[index]
                index = (index + 1) % len(self)
            else:
                if index == 0:
                    self.lshift(1)
                else:
                    self[index], self[index -
                                      1] = self[index - 1], self[index]
                index = (index - 1) % len(self)

    def __setitem__(self, k: int, v: object) -> None:
        if type(k) == int:
            if k > len(self):
                self.extend([None for _ in range(k - len(self) + 1)])
                super(PyshaList, self).__setitem__(k, v)
            else:
                super(PyshaList, self).__setitem__(k, v)
        else:
            raise TypeError("Type Of Index Must Be int.")

    # List Operators + Some Beautiful Things :) #

    def __add__(self, o: object) -> list:
        if type(o) == list:
            cp = self.copy()
            cp.extend(o)
            return cp
        else:
            tar = []
            for i in self:
                try:
                    tar.append(i + o)
                except:
                    tar.append(str(i) + " " + str(o))
            return tar

    def __iadd__(self, o: object) -> list:
        if type(o) == list:
            self.extend(o)
        else:
            for i in self:
                try:
                    self.append(i + o)
                except:
                    self.append(str(i) + " " + str(o))

    def __sub__(self, o: object) -> list:
        if type(o) == list:
            tar = []
            for i in self:
                if not i in o:
                    tar.append(i)
            return tar
        if type(o) == int or type(o) == float:
            tar = []
            for i in self:
                try:
                    tar.append(i - o)
                except:
                    pass
            return tar

    # **
    def __pow__(self, o: object) -> list:
        return list(map(lambda elem: elem ** o, self))

    # /
    def __truediv__(self, o: object) -> list:
        return list(map(lambda elem: elem/o, self))

    # //
    def __floordiv__(self, o: object) -> list:
        return list(map(lambda elem: elem//o, self))

    # <<
    def __lshift__(self, o: object) -> list:
        if type(o) == int:
            return self.lshift(o)
        elif type(o) == dict:
            for i, j in o.items():
                if type(i) == int:
                    self.shift_index(i, True, j)
                elif type(i) == tuple or type(i) == list or type(i) == set:
                    for a in i:
                        self.shift_index(a, True, j)
                else:
                    raise TypeError("Type Must Be int,tuple,list or set.")
        else:
            raise TypeError("Type Must Be int Or dict.")

    # >>
    def __rshift__(self, o: object) -> list:
        if type(o) == int:
            return self.rshift(o)
        elif type(o) == dict:
            for i, j in o.items():
                if type(i) == int:
                    self.shift_index(i, False, j)
                elif type(i) == tuple or type(i) == list or type(i) == set:
                    for a in i:
                        self.shift_index(a, False, j)
                else:
                    raise TypeError("Type Must Be int,tuple,list or set.")
        else:
            raise TypeError("Type Must Be int Or dict.")

    # %
    def __mod__(self, o: object) -> list:
        return list(map(lambda elem: elem % o, self))


"""
In PyshaString You Have Lots Of Methods To Do Some Things.
"""


class PyshaString(str):
    def inside(self, data_to_search: ListOrDict, search_mode="key", return_type=bool):
        """
        In This Function You Can Search That If A String Exists In Dict ( Or Inside Of All Elements ).
        Or If A String Exists In List And All Of Its Elements.

        Args :
            data_to_search ->  Data That Wanted To Search In That. For Example
                                Search In List : ["test", "some", "Hello World"] Or
                                Search In dict : {"test": "Hello World", "some": "Hello World"} 
                                -> ( search_mode must be "value" or "both" in dict input )
                                if string is Hello It Will Return True !

            search_mode ->  If You Passed A Dictionary You Can Choose Between "key" Or "value"
                            Or "all"/"both" To Search In Both Of Them. It Does Not Do Anything If
                            You Passed A List.

            return_type ->  If You Passed A Dictionary You Can Choose Between bool and dict or ...
                            if you pass bool, it will return True or False If It Finds Your String.
                            if you pass dict, it will return a dict with key and value of found element.
                            if you pass pass anything else, it will return oposite of search_mode. For
                            Example If You passed "key" And You Found Your String, It Will Return "value".

        """
        if type(data_to_search) == list:
            if self in data_to_search:
                return True
            for i in data_to_search:
                if self in i:
                    return True
            return False
        elif type(data_to_search) == dict:
            ans = {}
            if return_type == dict:
                ans = {}
            else:
                ans = []
            for i, j in data_to_search.items():
                if search_mode == "key":
                    if self in i:
                        if return_type == bool:
                            return True
                        elif return_type == dict:
                            ans[i] = j
                        else:
                            ans.append(j)
                elif search_mode == "value":
                    if self in j:
                        if return_type == bool:
                            return True
                        elif return_type == dict:
                            ans[i] = j
                        else:
                            ans.append(j)
                else:
                    if self in j or self in i:
                        if return_type == bool:
                            return True
                        elif return_type == dict:
                            ans[i] = j
                        else:
                            ans.append(j)
            if return_type != bool:
                return ans
            return False
        else:
            raise TypeError("Type Must Be list Or dict.")

    """
        In This Function You Can Search That an String Contains Some Other String From A List.

        Args :
            data_to_search ->  Strings That You Want To Search For. For Example
                                Search In List : [" ", ":", ","] And If Your String Is 
                                "Hello, Im Arshia" It will Return 5 Thats Index Of ",".
                                If Your String Is "Arshia:Hey" It Will Return 6 Thats Index Of ":".

            return_type ->  You Can pass int To Return Index Of Found String. Or Pass Other
                            Things Like str To Return Found String.

    """

    def find_first(self, data_to_search: list, return_type=int):
        for c, i in enumerate(self):
            if i in data_to_search:
                if return_type == int:
                    return c
                else:
                    return i

    """
        This Function Is Exactly find_first But It Will Return Last Index Of Found String.

        Args :
            data_to_search ->  Strings That You Want To Search For. For Example
                                Search In List : [" ", ":", ","] And If Your String Is 
                                "Hello, Im Arshia" It will Return 10 Thats Index Of " ".
                                If Your String Is "Arshia:Hey" It Will Return 6 Thats Index Of ":".

            return_type ->  You Can pass int To Return Index Of Found String. Or Pass Other
                            Things Like str To Return Found String.

    """

    def find_last(self, data_to_search: list, return_type=int):
        for c, i in enumerate(self[::-1]):
            if i in data_to_search:
                if return_type == int:
                    return len(self) - c - 1
                else:
                    return i

    """
        In This Function You Can Replace An Array With Another Array ! ( in a string )

        Args :
            data_to_replace ->  Data That You Wanted To Be Replaced For Example If Data Is 
                                ["Hello", "World"] And replace_with Argument Is ["Hi", "There"]
                                All Of "Hello" And "World" Will Be Replaced With "Hi" And "There".

            replace_with ->  Described In data_to_replace Argument.

    """

    def replace_array(self, data_to_replace: list, replace_with: list):
        for c, i in enumerate(data_to_replace):
            self = self.replace(i, replace_with[c])
        return self

    """
        This Is replace_array But You Can Pass Dictionary To Replace With.

        Args :
            data_to_replace ->  A Dictionary Of Data To Replace. For Example 
                                {"Hello": "Hi", "World": "There"} Will Replace "Hello" With "Hi"
                                And "World" With "There". ( All Of String )

    """

    def replace_dict(self, data_to_replace: dict):
        for j, i in data_to_replace.items():
            self = self.replace(j, i)
        return self

    """
        This Function Will Check If A List Of Strings Exists In Another String. In Specific
        Part Of It Or Not. It Will Get a Part Of String And Will Check It With A List Of Strings.

        Args :
            data_to_replace ->  A List Of Data That You Want To Check If Exists In Another String.
                                
            line_start ->  Line Start Number ( Starts With 0 )

            line_end ->  Line End Number ( Starts With 0 )

            start_index ->  Start Index Number ( Starts With 0 )

            end_index ->  End Index Number ( Starts With 0 )

    """

    def exists(self, data_to_search: list, line_start=-1, line_end=-1, start_index=-1, end_index=-1):
        if line_start == -1:
            line_start = 0
        if line_end == -1:
            line_end = len(self)
        if start_index == -1:
            start_index = 0
        if end_index == -1:
            end_index = len(self)
        data_to_search = '\n'.join(
            data_to_search.split('\n')[line_start:line_end])[start_index:end_index]
        for i in data_to_search:
            if i in self:
                return True
        return False

    """
        This Function Will Replace Any String That You Want But You Can Define Some Parts
        Not To Replace With Define Escapes !

        Args :
            string_to_find ->  String That You Want To Search For In String.
                                
            to_replace ->  String That You Want To Be Replaced With string_to_find.

            escape_char ->  Escape Character That You Want To Define. ( Default Is \ Character )

            remove_escapes ->  You Can Define That After Replacing, Escape Characters Should Remove
                                Or Not. ( Default Is False )
    """

    def replace_with_escape(self, string_to_find, to_replace, escape_char='\\', remove_escapes=False):
        last_index = self.find(string_to_find, 0)
        while last_index != -1:
            if self[last_index - 1] == escape_char:
                if remove_escapes:
                    self = self[:last_index-1] + self[last_index:]
                last_index = self.find(string_to_find, last_index)
                continue
            self = self[:last_index] + to_replace + \
                self[last_index + len(string_to_find):]
            last_index = self.find(string_to_find, last_index + 1)
        return self

    """
        This Function Is "<<=" Operator. It Will Append Another String To The End Of This String.
    """

    def __ilshift__(self, other):
        self = self + other
        return self

    """
        This Function Is "+=" Operator. It Will Append Another String To The End Of This String.
    """

    def __iadd__(self, other):
        self = self + other
        return self

    """
        This Function Is ">>=" Operator. It Will Append Another String To Start Of This String.
    """

    def __irshift__(self, other):
        self = other + self
        return self

    """
        This Function Is "<<" Operator. It Will Append Another String To The End Of This String. ( Not Set )
    """

    def __lshift__(self, other):
        return self + other

    """
        This Function Is ">>" Operator. It Will Append Another String To Start Of This String. ( Not Set )
    """

    def __rshift__(self, other):
        return other + self

    """
        This Function Is "-=" Operator. It Will Remove A String From This String.
    """

    def __isub__(self, other):
        self = self.replace(other, "")
        return self

    """
        This Function Is "-" Operator. It Will Remove A String From This String. ( Not Set )
    """

    def __sub__(self, other):
        return self.replace(other, "")


class PyshaDict(dict):
    @property
    def inverse(self):
        return PyshaDict({v: k for k, v in self.items()})

    @inverse.setter
    def inverse(self, value):
        value = {v: k for k, v in value.items()}
        self.update(value)

    def __invert__(self):
        return self.inverse

    def __isub__(self, other):
        del self[other]
        return self

    def __sub__(self, other):
        ans = self.copy()
        del ans[other]
        return ans



import platform
from .data_structures import PyshaDict
from .functions import *
from os import environ, popen, remove

"""
    This Class Will Be Use For Fun Part Of This Framework :)
    Most Important Part Of This Framework Exists In This Class.
"""


class command:
    def __init__(self):
        self.arr_data = []
        self.dic_tar = {}
        self._mode = 'p'

    @property
    def mode(self):
        return self._mode

    @mode.setter
    def mode(self, m):
        if m not in ['p', 'i']:
            raise ModeError(
                "Modes Should Be 'p' Or 'i'. Default Is 'p' / Print Mode.")
        else:
            self._mode = m

    """
    2 Types Of Loop Currently Exists. i(Input) and p(Print) Mode.
    In print mode you can make a print loop like that :
        <i:3>{You Are In _i_ Loop}
    And This Will Be Printed Like That : 
        You Are In 0 Loop
        You Are In 1 Loop
        You Are In 2 Loop

    And In Input Mode You Can Do This :
        <i:3>{Enter _i_ Number : }
    It Will Get 3 Input Like This :
        Enter 0 Number : (input0)
        Enter 1 Number : (input1)
        Enter 2 Number : (input2)
    And Will Return [input0,input1,input2]

    More Options :
        <i:3>{Enter _i++_ Number : }
        <i:3>{Enter _i--_ Number : }
        <i:2,j:2>{Enter Row _i++_ And Column _j++_  Number : }
        <i:2>{Enter (Fore.CYAN)[_i++_] Number : }
    
    Notes :
        Just Maded Maximum For 2 Variables Not More.
    """

    def loop(self, cmd, mode='p', c=True):
        if self.mode != 'p':
            mode = self.mode
        if mode == "p":  # Print Mode
            self.dic_tar.clear()
            self.arr_data.clear()
            alls = cmd.split(">")[0][1:]
            string = cmd.split("{")[1][:-1]
            for i in alls.split(','):
                itemp = i.split(':')
                self.arr_data.append(itemp[0].strip())
                self.dic_tar[itemp[0].strip()] = int(itemp[1].strip())
            if len(self.dic_tar.keys()) >= 4:
                raise LoopError(
                    "Loops With 4 Or More Variables Are Not Supported In This Version.")
            tars = make_possibles(self.dic_tar)
            for i in tars:
                temp = {}
                if not (type(i) == int) and len(i) == 2:
                    temp = {self.arr_data[0]: i[0], self.arr_data[1]: i[1]}
                elif not (type(i) == int) and len(self.dic_tar) == 3:
                    temp = {
                        self.arr_data[0]: i[0], self.arr_data[1]: i[1], self.arr_data[2]: i[2]}
                else:
                    temp = {self.arr_data[0]: i}
                display(temp, string, c)
        elif mode == 'i':  # input Mode
            self.dic_tar.clear()
            self.arr_data.clear()
            alls = cmd.split(">")[0][1:]
            string = cmd.split("{")[1][:-1]
            for i in alls.split(','):
                itemp = i.split(':')
                self.arr_data.append(itemp[0].strip())
                self.dic_tar[itemp[0].strip()] = int(itemp[1].strip())
            if len(self.dic_tar.keys()) >= 4:
                raise LoopError(
                    "Loops With 4 Or More Variables Are Not Supported In This Version.")
            tars = make_possibles(self.dic_tar)
            x = 0
            y = 0
            z = 0
            ls = []
            if len(self.dic_tar) == 2:
                x = (list(self.dic_tar.values())[0])
                y = (list(self.dic_tar.values())[1])
                ls = [["" for _ in range(y)] for __ in range(x)]
            elif len(self.dic_tar) == 1:
                x = (list(self.dic_tar.values())[0])
                ls = ["" for _ in range(x)]
            elif len(self.dic_tar) == 3:
                x = (list(self.dic_tar.values())[0])
                y = (list(self.dic_tar.values())[1])
                z = (list(self.dic_tar.values())[2])
                ls = [[["" for _ in range(z)] for __ in range(y)]
                      for ___ in range(x)]
            x = 0
            y = 0
            z = 0
            for i in tars:
                temp = {}
                if not (type(i) == int) and len(i) == 2:
                    temp = {self.arr_data[0]: i[0], self.arr_data[1]: i[1]}
                    try:
                        ls[x][y] = get_ans(temp, string, c)
                        ls[x][y + 1]
                        y += 1
                    except:
                        try:
                            ls[x + 1][y]
                            x += 1
                            y = 0
                        except:
                            pass
                elif not (type(i) == int) and len(i) == 3:
                    temp = {
                        self.arr_data[0]: i[0], self.arr_data[1]: i[1], self.arr_data[2]: i[2]}
                    try:
                        ls[x][y][z] = get_ans(temp, string, c)
                        ls[x][y][z + 1]
                        z += 1
                    except:
                        try:
                            ls[x][y + 1][z]
                            y += 1
                            z = 0
                        except:
                            try:
                                ls[x + 1][y][z]
                                x += 1
                                y = 0
                                z = 0
                            except:
                                pass
                else:
                    temp = {self.arr_data[0]: i}
                    try:
                        ls[x] = get_ans(temp, string, c)
                        ls[x]
                        x += 1
                    except:
                        pass

            return ls
        else:
            raise ValueError("Mode Should Be 'p' Or 'i' !")

    """
    This Function Act As c++ One Line Condition.
    You Can Use This Function Like :
        condition('i > j ? "i Is Grater" : "j Is Grater"',i=20,j=23)
    Output Will Be :
        "j Is Grater"
    
    Notes & Warnings :
        -> '(Quotes) Are Not Working, Always Use "(Double Quotes)
    """

    def condition(self, check, p=True, **kwargs):
        if not re.match("""(\"?)[ -~]+(\"?)(>|<|==|>=|<=)(\"?)[ -~]+(\"?)\?(\"?)[ -~]+(\"?):(\"?)[ -~]+(\"?)""", ''.join(check.split(' '))):
            raise ConditionError("An Error In Parsing Condition :(")
        if "eval" in check:
            if not check[check.find("eval") - 1:].startswith("\eval"):
                raise SecurityError("Using This Funcion Is Forbidden.")
        if "exec" in check:
            if not check[check.find("exec") - 1:].startswith("\exec"):
                raise SecurityError("Using This Funcion Is Forbidden.")
        index_of_ok_sign = not_between_index("?", check)
        p1 = index_split(check, index_of_ok_sign)[0].strip()
        p2 = index_split(check, index_of_ok_sign)[1].strip()
        ans = eval(p1, kwargs, {})
        index_of_ok_sign = not_between_index(":", p2)
        ok = index_split(p2, index_of_ok_sign)[0].strip()
        not_ok = index_split(p2, index_of_ok_sign)[1].strip()
        if ans:
            if p:
                print(eval(ok, kwargs, {}))
            else:
                return eval(ok, kwargs, {})
        else:
            if p:
                print(eval(not_ok, kwargs, {}))
            else:
                return eval(not_ok, kwargs, {})

    """
        In Progress But This Function Should Get Text Between () Of $(cmd)
        And Run It In Terminal Then Replace The Output With $(cmd) And Should
        Get #(cmd) Too And Run It As Python Code And Replace This Output Too.
            Problem Now :
                Use This 2 Thing Together.
        
        Args : 
            string          ->    The String Pattern That You Want To Make.

            priority        ->    If Its 0 First System Commands Will Replace Then
                                Python Commands Will Replace. And If Its 1 First
                                Python Commands Will Replace Then System Commands
                                Will Be Replaced. Default Is 0.
            
            strip_response  ->     Default Is True. It Wanted To Know That Your
                                    String That Replaced Should Be striped ? ( Remove
                                    Addtion New Lines And Spaces In Start And End Of 
                                    String ? )
        
        Example :
            exe("You Current Directory :\n $(dir)\nFinished :)")
    """

    def exe(self, string, priority=0, strip_response=True):
        s = re.compile(
            "\$\([A-Za-z0-9_ \$\&\!\@\#\%\^\*\+\=\-\]\[\~\?\'\"\;\:]+\)")
        p = re.compile(
            "\#\([A-Za-z0-9_ \$\&\!\@\#\%\^\*\+\=\-\]\[\~\?\'\"\;\:]+\)")
        if priority == 0:
            # First System Then Python
            for cmd in s.finditer(string):
                cmds = cmd.group()[2:-1]
                ans = popen(cmds).read()
                string = string[:cmd.start()] + ans + \
                    string[cmd.start() + len(cmd.group()):]
            for cmd in p.finditer(string):
                cmds = cmd.group()[2:-1]
                f = open("temp.py", "w")
                f.write(cmds)
                f.close()
                ans = popen(environ.get("python") + " temp.py").read()
                remove("temp.py")
                string = string[:cmd.start()] + ans + \
                    string[cmd.start() + len(cmd.group()):]
            if strip_response:
                return string.strip()
            else:
                return string
        elif priority == 1:
            # First Python Then System
            for cmd in p.finditer(string):
                cmds = cmd.group()[2:-1]
                f = open("temp.py", "w")
                f.write(cmds)
                f.close()
                ans = popen(environ.get("python") + " temp.py").read()
                remove("temp.py")
                string = string[:cmd.start()] + ans + \
                    string[cmd.start() + len(cmd.group()):]
            for cmd in s.finditer(string):
                cmds = cmd.group()[2:-1]
                ans = popen(cmds).read()
                string = string[:cmd.start()] + ans + \
                    string[cmd.start() + len(cmd.group()):]
            if strip_response:
                return string.strip()
            else:
                return string
        else:
            ValueError(
                "Priority Should Be 0 (System First) Or 1 (Python First).")


"""
    The Famous Switch-Case :)
"""


class Switch:
    def __init__(self, var):
        self.var = var

    def case(self, tar, func, args=tuple()):
        if self.var == tar or (hasattr(tar, '__name__') and tar.__name__ == "Case" and tar == self.var) or (tar == self.var) or (hasattr(tar, '__name__') and tar.__name__ == "Default"):
            func(*args)

    def __getitem__(self, all_funcs):
        dic = {}
        for i in all_funcs:
            dic[i.start] = i.stop
        for i, j in dic.items():
            if self.var == i or (hasattr(i, '__name__') and i.__name__ == "Case" and i.tar == self.var) or (hasattr(i, 'tar') and i.tar == self.var) or (hasattr(i, '__name__') and i.__name__ == "Default"):
                if hasattr(j, '__call__'):
                    j()
                    return
                elif type(j) == str:
                    parts = j.split('\n')
                    if parts[0].strip():
                        t = parts[0].index(j.split('\n')[0].strip())
                    else:
                        t = parts[1].index(j.split('\n')[1].strip())
                    ok_code = ""
                    for i in j.split('\n'):
                        ok_code += i[t:] + "\n"
                    b = compile(ok_code, 'user_code', 'exec')
                    a = exec(b)
                    if a != None:
                        print(a)
                    return
                if len(j) == 3:
                    j[0](*j[1], **j[2])
                    return
                elif len(j) == 2:
                    j[0](*j[1])
                    return
                elif len(j) == 1:
                    j[0]()
                    return

    def cases(self, all_funcs):
        for i, j in all_funcs.items():
            if self.var == i or (hasattr(i, '__name__') and i.__name__ == "Case" and i.tar == self.var) or (hasattr(i, 'tar') and i.tar == self.var) or (hasattr(i, '__name__') and i.__name__ == "Default"):
                if hasattr(j, '__call__'):
                    j()
                    return
                elif type(j) == str:
                    parts = j.split('\n')
                    if parts[0].strip():
                        t = parts[0].index(j.split('\n')[0].strip())
                    else:
                        t = parts[1].index(j.split('\n')[1].strip())
                    ok_code = ""
                    for i in j.split('\n'):
                        ok_code += i[t:] + "\n"
                    b = compile(ok_code, 'user_code', 'exec')
                    a = exec(b)
                    if a != None:
                        print(a)
                    return
                if len(j) == 3:
                    j[0](*j[1], **j[2])
                    return
                elif len(j) == 2:
                    j[0](*j[1])
                    return
                elif len(j) == 1:
                    j[0]()
                    return


"""
    Default Case Of Switch-Case
"""


class Default:
    pass


"""
    Define Every Case
"""


class Case:
    def __init__(self, tar):
        self.tar = tar


"""
    Define Variables Of Switch-Case ( Just lambda Case )
"""


class Vars:
    pass


"""
    Pretty Percent Printer For Loading Some Progress Or Downloading :)
"""


class PercentPrinter:
    def __init__(self, chars=100, pass_color=fore["reset"], loading_color=fore["reset"]):
        self._percent = 0
        self.chars = chars
        self.pass_color = pass_color
        self.loading_color = loading_color

    def show(self, char_ok='#', char_loading='-', end='\n'):
        num = self._percent
        string = "\r"
        string += (self.pass_color + char_ok + Fore.RESET) * \
            int(num * (self.chars / 100))
        string += (self.loading_color + char_loading + Fore.RESET) * \
            int((100 - num) * (self.chars / 100))
        print(string, end=end)

    @property
    def percent(self):
        return self._percent

    @percent.setter
    def percent(self, p):
        if p > 100:
            raise ValueError("Value Most Be Less Than 100 !")
        elif p < 0:
            raise ValueError("Value Most Be More Than 0 !")
        else:
            self._percent = p

    def increase(self, p=1, show=True):
        self._percent += p
        if show:
            self.show()

    def finish(self, show=True):
        self._percent = 100
        if show:
            self.show()


"""
This Class Will Make An Environment For You To Have Easier And Better Experience :)
"""


class CrossPlatformer:
    def __init__(self, supported_os=["linux", "windows", "mac"]) -> None:
        self.user_os = platform.system().strip().lower()
        self.all_commands = {}
        self.supported_os = supported_os
        self.platforms = {
            "windows": ["windows", "win32"],
            "linux": ["linux", "linux2"],
            "mac": ["mac", "darwin"]
        }

    def get_os_name(self):
        """Get User Os Name"""
        if self.user_os == "linux" or self.user_os == "linux2":
            return "Linux"
        elif self.user_os == "darwin":
            return "Mac"
        elif self.user_os == "win32" or self.user_os == "windows":
            return "Windows"
        else:
            return "Unknown"

    def is_os_supported(self, os_name):
        """Check If An Os Is Supported"""
        for i, j in self.platforms.items():
            if os_name.lower().strip() in j or os_name.lower().strip() == i:
                return True
        return False

    def get_os_version(self):
        """Get User Os Version"""
        if self.user_os == "linux" or self.user_os == "linux2":
            return platform.linux_distribution()[0]
        elif self.user_os == "darwin":
            return platform.mac_ver()[0]
        elif self.user_os == "win32" or self.user_os == "windows":
            return platform.win32_ver()[0]
        else:
            return "Unknown"

    def get_os_command(self, name):
        """
        Get Command For User Os. Arguments :
            name : Command Name
        For Example :
            If You Set add_os_commands("clear", {"linux": "clear", "windows": "cls", "mac": "clear"}) :
                get_os_command("clear")
            Will Return You The Command Depends On Your Os.
        """
        try:
            return self.all_commands[name][self.user_os]
        except:
            return None

    def add_os_commands(self, name, commands: dict):
        """
        Set Command For User Os. Arguments Are : 
            name : Name Of Command
            commands : Dictionary Of Commands
        For Example : 
            add_os_commands("git", {"linux": "git", "windows": "git.exe"})
        Or : 
            add_os_commands("clear", {"linux": "clear", "windows": "cls", "mac": "clear"})
        """
        for i in commands:
            if not self.is_os_supported(i):
                raise ValueError(f"This OS ({i}) Is Not Supported !")
        self.all_commands[name] = commands

    def __getitem__(self, name):
        return self.get_os_command(name)

    def __setitem__(self, name, commands):
        self.add_os_commands(name, commands)



class LoopError(Exception):
    pass

class ModeError(Exception):
    pass

class ConditionError(Exception):
    pass

class StackError(Exception):
    pass

class QueueError(Exception):
    pass

class SecurityError(Exception):
    pass

class AlgorithmError(Exception):
    pass

def exception_create(name):
    return type(name,(Exception,),{})


import pyfiglet
from .classes import *
from .data_structures import *
from .auth import *
from .enums import *
from .decorators import *

"""
Abbreviation of Pretty Print.
Args:
    You Can Give As Much As You Can For Print :)

Kwargs:
    curly_c : Set Curly Brackets Colors or "{}" Color.
    colon_c : Set Colon Colors or ":" Color.
    quote_c : Set Quote Colors or "'" Color.
    mode    : Set Mode Of pp. Modes Are 'k' And 'i'. You Can Read About
                Them In Modes Part.
    end     : This Will Set End Of Print. Default Is '\n'.

Modes : 
    k : This Will Use all colors That You Specified As Kwargs.
    i : This Will Ignore Colors(curly_c,colon_c,quote_c) That You
            Specified In Kwargs. 
"""


def pp(*args, curly_c=Fore.RESET, colon_c=Fore.RESET, quote_c=Fore.RESET, mode='k', end='\n'):
    for i in args:
        if type(i) == list:
            print("[", ' , '.join([str(j) for j in i]), "]", end=end)
        elif type(i) == dict:
            print(curly_c + "{" + Fore.RESET)
            for a, b in i.items():
                print("\t" + quote_c + "'" + Fore.RESET + a.replace("'", "\\" + quote_c + "'" + Fore.RESET) + quote_c + "'" + Fore.RESET + colon_c +
                      " : " + Fore.RESET + quote_c + "'" + Fore.RESET + b.replace("'", "\\" + quote_c + "'" + Fore.RESET) + quote_c + "'" + Fore.RESET + " ,")
            print(curly_c + "}" + Fore.RESET, end=end)
        elif type(i) == set:
            print(curly_c + "{" + Fore.RESET)
            for a in i:
                print("\t", a, ",")
            print(curly_c + "}" + Fore.RESET, end=end)
        elif type(i) == str:
            if mode == 'k':  # Kwargs Mode
                i = i.replace("{", curly_c + "{" + Fore.RESET)
                i = i.replace("}", curly_c + "}" + Fore.RESET)
                i = i.replace("'", quote_c + "'" + Fore.RESET)
                i = colorize(i)
                i = i.replace(":", colon_c + ":" + Fore.RESET)
                print(i, end=end)
            elif mode == 'i':  # Input Mode
                i = colorize(i)
                print(i, end=end)
        else:
            print(i, end=end)


"""
This Function Will Draw A Rectangle Plus Your Target Text Middle Of It.
Args:
    Just Your Texts. You Can Pass It Like rect("hello","Thanks") Or rect("hello\nThanks")
kwargs:
    text_color    ->   Color Of Your Text.

    first_line    ->   This Parameter Will Set First Line Character And Color.
                        You Can Pass A Character, Default Text Color Will Be Setted.
                        Or You Can Pass A Tuple That Contains Character And Color Like
                        ('-',Fore.RED).

    sep           ->   This Parameter Will Set Corners Character And Color.
                        You Can Pass A Character, Default Text Color Will Be Setted.
                        Or You Can Pass A Tuple That Contains Character And Color Like
                        ('-',Fore.RED).

    last_line     ->   This Parameter Will Set Last Line Character And Color.
                        You Can Pass A Character, Default Text Color Will Be Setted.
                        Or You Can Pass A Tuple That Contains Character And Color Like
                        ('-',Fore.RED).

    distance_up   ->   This Parameter Will Set Distance Between Text And First Line.
                        You Can Pass An Integer That Is Line Numbers Of Distance.

    distance_down ->   This Parameter Will Set Distance Between Text And Last Line.
                        You Can Pass An Integer That Is Line Numbers Of Distance.

    length        ->   This Parameter Will Set Number Of Characters In First/Last
                        Line. You Can Pass An Integer That Is Character Count Of First/Last Line.

    p            ->   This Parameter Will Set Print Mode. If Its True It Will Print
                        And If Its False It Will Return Rectangle As String.


Example : 
    * Code
        rect("Hello\nWelcome To Pysha :)",first_line='-',last_line='-',distance_up=2,distance_down=2)
    
    * Output
        ------------------------------
        |                            |
        |                            |
        |           Hello            |
        |    Welcome To Pysha :)     |
        |                            |
        |                            |
        ------------------------------
"""


def rect(*args, text_color=Fore.RESET, first_line=("=", Fore.RESET), sep=("|", Fore.RESET), last_line=("=", Fore.RESET), distance_up=1, distance_down=1, length=30, p=True):
    ans = []
    for i in args:
        ans.extend(i.split('\n'))
    args = ans.copy()
    del ans
    if len(first_line) == 1:
        first_line = first_line, Fore.RESET
    if len(sep) == 1:
        sep = sep, Fore.RESET
    if len(last_line) == 1:
        last_line = last_line, Fore.RESET
    max_length = 0
    for i in args:
        if len(i) > max_length:
            max_length = len(i)
    max_length += 10
    if length > max_length:
        max_length = length
    if p:
        print(first_line[1] + first_line[0] * max_length + Fore.RESET)
        for i in range(distance_up):
            print(sep[1] + sep[0] + Fore.RESET, end='')
            print(' ' * (max_length-2), end='')
            print(sep[1] + sep[0] + Fore.RESET)
        for i in args:
            print(sep[1] + sep[0] + Fore.RESET, end='')
            print(text_color + i.center(max_length-2) + Fore.RESET, end='')
            print(sep[1] + sep[0] + Fore.RESET)
        for i in range(distance_down):
            print(sep[1] + sep[0] + Fore.RESET, end='')
            print(' ' * (max_length-2), end='')
            print(sep[1] + sep[0] + Fore.RESET)
        print(last_line[1] + last_line[0] * max_length + Fore.RESET)
    else:
        s = first_line[1] + first_line[0] * max_length + Fore.RESET
        for i in range(distance_up):
            s += sep[1] + sep[0] + Fore.RESET + ' ' * (max_length-2)
            s += sep[1] + sep[0] + Fore.RESET + "\n"
        for i in args:
            s += (sep[1] + sep[0] + Fore.RESET)
            s += (text_color + i.center(max_length-2) + Fore.RESET)
            s += (sep[1] + sep[0] + Fore.RESET + "\n")
        for i in range(distance_down):
            s += (sep[1] + sep[0] + Fore.RESET) + (' ' * (max_length-2))
            s += (sep[1] + sep[0] + Fore.RESET + "\n")
        s += (last_line[1] + last_line[0] * max_length + Fore.RESET)
        return s


"""
This Function Directly Use pyfiglet Library For Making A Banner.
Args:
    text -> The Text That You Want To Convert To A Banner.

    font -> Target Font From Figlet. If You Pass Empty Font Will Be Default
              Figlet Font.

    p    -> This Parameter Will Set Print Mode. If Its True It Will Print
              And If Its False It Will Return Banner As String.

Example : 
    * Code
        banner("pysha","chunky")
    
    * Output
        ______               __
        |   __ \.--.--.-----.|  |--.---.-.
        |    __/|  |  |__ --||     |  _  |
        |___|   |___  |_____||__|__|___._|
                |_____|
"""


def banner(text, font="", p=True):
    if font:
        try:
            if p:
                print(pyfiglet.figlet_format(text, font=font))
            else:
                return pyfiglet.figlet_format(text, font=font)
        except pyfiglet.FontNotFound:
            print("Font " + font + " Not Found :(")
        except:
            raise
        return
    if p:
        print(pyfiglet.figlet_format(text))
    else:
        return pyfiglet.figlet_format(text)


"""
This Function Will Draw A Line In Terminal.
Kwargs:
    char -> You Can Pass (char='-') Then Color Will Be Normal Color. Or
            You Can Pass (char=('-',Fore.[color])) Then Color Will Be [color].
            Default Is ('=',Fore.RESET)

    count -> Count Of Characters That Line Have.

    p     -> This Parameter Will Set Print Mode. If Its True It Will Print
              And If Its False It Will Return Line As String.
"""


def l(char=('=', Fore.RESET), count=30, p=True):
    if type(char) == str or ((type(char) == list or type(char) == set or type(char) == tuple) and len(char) == 1):
        char = char, Fore.RESET
    if p:
        print(char[1] + char[0] * count + Fore.RESET)
    else:
        return char[1] + char[0] * count + Fore.RESET


"""
This Function Will pp The Arguments.
You Can Set A Prompt If You Want. ( Prompt Will Not Be pp )
After pp And Prompt It Will Get Input And Returns Input.
You Can Change Prompt Color By Pass (prompt,Fore.[color]) To Prompt.
Kwargs:
    prompt -> You Can Pass (prompt='Enter Your Name : ') And It Will Act As
              input('Enter Your Name : '). Or You Can Pass (prompt=('Name : ',Fore.RED))
              And It Will Print 'Name : ' In Red Color.
Example : 
    * Code
        name = xp("(Fore.RED)[W3LC0M3] To (Fore.GREEN)[T3ST] Application.",prompt=('Name : ',Fore.CYAN))
        pp(name)
    * Output
        _RED ->|W3LC0M3|_ To _GREEN -> |T3ST|_ Application.
        _CYAN -> |Name :|_ (input)
        (input)
"""


def xp(*args, prompt=('', Fore.RESET)):
    if (len(prompt) == 1 and (type(prompt) == tuple or type(prompt) == list or type(prompt) == set)) or type(prompt) == str:
        prompt = prompt, Fore.RESET
    pp(*args, mode='i', end='')
    return input(prompt[1] + prompt[0] + Fore.RESET)



import base64
import string
from .exceptions import *
from .enums import Algorithms
from pickle import Unpickler, Pickler

"""
Crypto Class Is Made For Do Encryption And Decryption More Coder Friendly :)
"""


class Crypto:
    def __init__(self, enc_ls, dec_ls, key):
        self.enc_ls = enc_ls
        self.dec_ls = dec_ls
        self.key = key

    def __call__(self, s):
        t_enc = s
        for i in self.enc_ls:
            t_enc = i(t_enc, self.key)
        return t_enc

    def __eq__(self, o) -> bool:
        if len(o.enc_ls) == len(self.enc_ls):
            return True
        return False

    def __gt__(self, o) -> bool:
        if len(o.enc_ls) < len(self.enc_ls):
            return True
        return False

    def __ge__(self, o) -> bool:
        if len(o.enc_ls) <= len(self.enc_ls):
            return True
        return False

    def __ne__(self, o) -> bool:
        if len(o.enc_ls) != len(self.enc_ls):
            return True
        return False

    def __lt__(self, o) -> bool:
        if len(o.enc_ls) > len(self.enc_ls):
            return True
        return False

    def __le__(self, o) -> bool:
        if len(o.enc_ls) >= len(self.enc_ls):
            return True
        return False

    def enc(self, s):
        t_enc = s
        for i in self.enc_ls:
            t_enc = i(t_enc, self.key)
        return t_enc

    def dec(self, s):
        t_dec = s
        for i in self.dec_ls:
            t_dec = i(t_dec, self.key)
        return t_dec

    def __str__(self):
        return '('.join([str(i).split(' ')[1] for i in self.enc_ls]) + f"(string)) -> Key Is {self.key}"


"""
A Class To Save Some Variables With Some Simple Encryption.
"""


class Save:
    def __init__(self, file_name, **kwargs):
        self.file_name = file_name
        self.vars_list = kwargs

    def add_var(self, key, value):
        self.vars_list[key] = value

    def add_vars(self, **kwargs):
        self.vars_list.update(kwargs)

    def __call__(self):
        Pickler(open(self.file_name, "wb")).dump(self.vars_list)

    def save(self):
        Pickler(open(self.file_name, "wb")).dump(self.vars_list)


"""
A Class To Load Saved Variables
"""


class Load:
    def __init__(self, file_name):
        self.file_name = file_name
        self.vars_list = Unpickler(open(self.file_name, "rb")).load()

    def vars(self):
        return self.vars_list

    def __call__(self, module):
        for i, j in self.vars_list.items():
            setattr(module, i, j)

    def load(self, module):
        for i, j in self.vars_list.items():
            setattr(module, i, j)


"""
xor function. got string And key At The End; It Will Encrypt/Decrypt Your
Data.
"""


def xor(string, key=0):
    if type(string) == str:
        string = bytes(string, encoding='utf-8')
    elif type(string) == bytes:
        pass
    else:
        raise TypeError("Type Should Be 'str' Or 'bytes'.")
    if type(key) == int:
        ans = ""
        for i in string:
            ans += chr(i ^ key)
    else:
        raise TypeError("key Type Must Be Int.")
    return ans


"""
base64 Encoder function. got string And key (Fake) At The End; It Will Encrypt Your
Data.
"""


def b64_en(string, k=""):
    if type(string) == str:
        string = bytes(string, encoding='utf-8')
    elif type(string) == bytes:
        pass
    else:
        raise TypeError("Type Should Be 'str' Or 'bytes'.")
    return base64.b64encode(string).decode('utf-8')


"""
base64 Decoder function. got string And key (Fake) At The End; It Will Decrypt Your
Data.
"""


def b64_de(string, k=""):
    if type(string) == str:
        string = bytes(string, encoding='utf-8')
    elif type(string) == bytes:
        pass
    else:
        raise TypeError("Type Should Be 'str' Or 'bytes'.")
    return base64.b64decode(string).decode('utf-8')


"""
base32 Encoder function. got string And key (Fake) At The End; It Will Encrypt Your
Data.
"""


def b32_en(string, k=""):
    if type(string) == str:
        string = bytes(string, encoding='utf-8')
    elif type(string) == bytes:
        pass
    else:
        raise TypeError("Type Should Be 'str' Or 'bytes'.")
    return base64.b32encode(string).decode('utf-8')


"""
base32 Decoder function. got string And key (Fake) At The End; It Will Decrypt Your
Data.
"""


def b32_de(string, k=""):
    if type(string) == str:
        string = bytes(string, encoding='utf-8')
    elif type(string) == bytes:
        pass
    else:
        raise TypeError("Type Should Be 'str' Or 'bytes'.")
    return base64.b32decode(string).decode('utf-8')


"""
base16 Encoder function. got string And key (Fake) At The End; It Will Encrypt Your
Data.
"""


def b16_en(string, k=""):
    if type(string) == str:
        string = bytes(string, encoding='utf-8')
    elif type(string) == bytes:
        pass
    else:
        raise TypeError("Type Should Be 'str' Or 'bytes'.")
    return base64.b16encode(string).decode('utf-8')


"""
base16 Decoder function. got string And key (Fake) At The End; It Will Decrypt Your
Data.
"""


def b16_de(string, k=""):
    if type(string) == str:
        string = bytes(string, encoding='utf-8')
    elif type(string) == bytes:
        pass
    else:
        raise TypeError("Type Should Be 'str' Or 'bytes'.")
    return base64.b16decode(string).decode('utf-8')


"""
base85 Encoder function. got string And key (Fake) At The End; It Will Encrypt Your
Data.
"""


def b85_en(string, k=""):
    if type(string) == str:
        string = bytes(string, encoding='utf-8')
    elif type(string) == bytes:
        pass
    else:
        raise TypeError("Type Should Be 'str' Or 'bytes'.")
    return base64.b85encode(string).decode('utf-8')


"""
base85 Decoder function. got string And key (Fake) At The End; It Will Decrypt Your
Data.
"""


def b85_de(string, k=""):
    if type(string) == str:
        string = bytes(string, encoding='utf-8')
    elif type(string) == bytes:
        pass
    else:
        raise TypeError("Type Should Be 'str' Or 'bytes'.")
    return base64.b85decode(string).decode('utf-8')


"""
Cipher Encoder function. got string And key At The End; It Will Encrypt Your
Data.
"""


def cipher_en(s, key=1):
    if type(key) == str:
        try:
            key = int(key)
        except:
            raise TypeError("Type Should Be 'int'.")
    elif type(key) == int:
        pass
    else:
        raise TypeError("Type Should Be 'int'.")

    slow = string.ascii_lowercase + string.ascii_lowercase
    sup = string.ascii_uppercase + string.ascii_uppercase

    ans = ""

    for i in s:
        if i in string.ascii_lowercase:
            ans += slow[slow.index(i) + key]
        elif i in string.ascii_uppercase:
            ans += sup[sup.index(i) + key]
        else:
            ans += i

    return ans


"""
Cipher Decoder function. got string And key At The End; It Will Decrypt Your
Data.
"""


def cipher_de(s, key=1):
    if type(key) == str:
        try:
            key = int(key)
        except:
            raise TypeError("Type Should Be 'int'.")
    elif type(key) == int:
        pass
    else:
        raise TypeError("Type Should Be 'int'.")

    slow = string.ascii_lowercase + string.ascii_lowercase
    sup = string.ascii_uppercase + string.ascii_uppercase

    ans = ""

    for i in s:
        if i in string.ascii_lowercase:
            ans += slow[slow.index(i) - key]
        elif i in string.ascii_uppercase:
            ans += sup[sup.index(i) - key]
        else:
            ans += i

    return ans


"""
ROT13 function. It Will Encrypt Your Data In This Algorythm.
"""


def rot13(string):
    slow = string.ascii_lowercase
    sup = string.ascii_uppercase

    ans = ""

    for i in string:
        if i in string.ascii_lowercase:
            ans += slow[(slow.index(i) + 13) % 26]
        elif i in string.ascii_uppercase:
            ans += sup[(sup.index(i) + 13) % 26]
        else:
            ans += i

    return ans


"""
ROT13 Decoder function. It Will Decrypt Your Data In This Algorythm.
"""


def unrot13(string):
    slow = string.ascii_lowercase
    sup = string.ascii_uppercase

    ans = ""

    for i in string:
        if i in string.ascii_lowercase:
            ans += slow[(slow.index(i) - 13) % 26]
        elif i in string.ascii_uppercase:
            ans += sup[(sup.index(i) - 13) % 26]
        else:
            ans += i

    return ans


"""
make_enc Function Will Make You An instance Of Crypto Class That Contains A
 Chain Of Encryption Algorithm. In Fact You Are Free To Use Just One Or More.
Args/Kwargs :
    alg     ->    The Algorithm That Can Be Like Algorithms.XOR Or Like 
                    [Algorithms.XOR,Algorithms.Base64]
    
    key     ->    The Target Key That You Want To Use In Encryption Process.

Example :

    a = make_enc([Algorithms.XOR,Algorithms.Base64],10)
    a.enc("Hello")    # Encrypt "Hello" Equals "Qm9mZmU="
    a.dec("Qm9mZmU=") # Decrypt "Qm9mZmU" Equals "Hello"
"""


def make_enc(alg, key=b""):
    if str(type(alg)) == "<enum 'Algorithms'>":
        if alg.name == Algorithms.XOR.name:
            return Crypto([xor], [xor], key)
        elif alg.name == Algorithms.ROT13.name:
            return Crypto([rot13], [unrot13], key)
        elif alg.name == Algorithms.Base64.name:
            return Crypto([b64_en], [b64_de], key)
        elif alg.name == Algorithms.Cipher.name:
            return Crypto([cipher_en], [cipher_de], key)
        elif alg.name == Algorithms.Base16.name:
            return Crypto([b16_en], [b16_de], key)
        elif alg.name == Algorithms.Base32.name:
            return Crypto([b32_en], [b32_de], key)
        elif alg.name == Algorithms.Base85.name:
            return Crypto([b85_en], [b85_de], key)
        else:
            raise AlgorithmError(
                f"This Algorithm Is Not Available. We Will Be Happy If You Help Us To Make It :)\nGithub : https://github.com/Arshiatmi/Pysha")
    elif type(alg) == list or type(alg) == set:
        ls_en = []
        ls_de = []
        for i in alg:
            if hasattr(i, 'name') and i.name == Algorithms.XOR.name:
                ls_en.append(xor)
                ls_de.append(xor)
            elif hasattr(i, 'name') and i.name == Algorithms.Base64.name:
                ls_en.append(b64_en)
                ls_de.append(b64_de)
            elif hasattr(i, 'name') and i.name == Algorithms.Cipher.name:
                ls_en.append(cipher_en)
                ls_de.append(cipher_de)
            elif hasattr(i, 'name') and i.name == Algorithms.Base16.name:
                ls_en.append(b16_en)
                ls_de.append(b16_de)
            elif hasattr(i, 'name') and i.name == Algorithms.Base32.name:
                ls_en.append(b32_en)
                ls_de.append(b32_de)
            elif hasattr(i, 'name') and i.name == Algorithms.Base85.name:
                ls_en.append(b85_en)
                ls_de.append(b85_de)
            elif hasattr(i, 'name') and i.name == Algorithms.ROT13.name:
                ls_en.append(rot13)
                ls_de.append(unrot13)
            else:
                raise AlgorithmError(
                    f"Type {i} Is Not Supported. Just (Algorithms.Base64/Algorithms.XOR/Algorithms.Cypher) Is Supported.")
        return Crypto(ls_en, ls_de[::-1], key)
    else:
        raise AlgorithmError(
            f"This Type Of Algorithm Is Not Supported. Just enum.EnumMeta (Algorithms.Base64/...),list and set Are Supported.")



import time
from typing import Callable, Union
from os import mkdir, makedirs, rmdir
from shutil import rmtree
from .colors import *
from .exceptions import *
import re
import readchar
import sys
from .security import *


"""
This Will Act Like Input. But Characters That User Enters Will Be Colored :)
Args :
    prompt      ->   The Text For Print And Wait For Input.
    
    out         ->   Just Standard Output. DO NOT TOUCH THIS :/

    char_color  ->   The Color Of Character. Each Character Will Have This 
                     Color.

"""


def colorprompt(prompt: str, out=sys.stdout, char_color=fore["reset"]):
    out.write(prompt)
    out.flush()
    password = ""
    while True:
        ch = str(readchar.readchar(), encoding='UTF-8')
        if ch == '\r':
            break
        elif ch == '\b':
            out.write('\b \b')
            password = password[0:len(password)-1]
            out.flush()
        elif ch == "\x03":
            print("\n")
            raise KeyboardInterrupt("You Pressed Ctrl+C")
        else:
            password += ch
            out.write(char_color + ch + fore["reset"])
            out.flush()
    print()
    return password


"""
This Will Get Password And Replace Characters With A Mask For Example '*'.
Args :
    prompt      ->   The Text For Print And Wait For Input.

    out         ->   Just Standard Output. DO NOT TOUCH THIS :/

    mask_color  ->   The Color Of Mask. For Example You Can Print Red Stars For
                     Each Character :))

    mask        ->   The Mask. Every Character That User Type, Will Be Replaced With
                     This Character :)
"""


def passprompt(prompt: str, out=sys.stdout, mask_color=fore["reset"], mask='*') -> str:
    out.write(prompt)
    out.flush()
    password = ""
    while True:
        ch = str(readchar.readchar(), encoding='UTF-8')
        if ch == '\r':
            break
        elif ch == '\b':
            out.write('\b \b')
            password = password[0:len(password)-1]
            out.flush()
        elif ch == "\x03":
            print("\n")
            raise KeyboardInterrupt("You Pressed Ctrl+C")
        else:
            password += ch
            out.write(mask_color + mask + fore["reset"])
            out.flush()
    print()
    return password


"""
This Function Will Help You To Recover Escaped Characters That Wrongly Splitted.
Args :
    target         ->   A List Of Splitted String/ A String That You Want To Escape.

    char           ->   The Character That You Want To Escape.

Example :
    escape_translator("Hello Lets Escape All < That Have \ Like \< . All \< Will Be Escaped.","<")

Out :
    ['Hello Lets Escape All ', ' That Have \\ Like \\< . All \\< Will Be Escaped.']
"""


def escape_translator(target, char: str) -> list:
    if type(target) == str:
        target = target.split(char)
    elif type(target) == list:
        pass
    else:
        raise TypeError("Type Should Be List Or String.")
    ans = []
    escape = False
    for i in target:
        print(i, ans)
        if i[-1] == "\\":
            if escape:
                ans[-1] = ans[-1] + char + i
            else:
                ans.append(i)
            escape = True
        else:
            if escape:
                ans[-1] = ans[-1] + char + i
            else:
                ans.append(i)
            escape = False
    return ans


"""
It Will Colorize The Input.
Example :
    colorize("(Fore.GREEN)[Hello !]")
Output :
    Green Foreground -> |Hello !|

Possible Colors:
    fore = {
        "green" : Fore.GREEN,
        "black":Fore.BLACK,
        "red":Fore.RED,
        "yellow":Fore.YELLOW,
        "blue":Fore.BLUE,
        "cyan":Fore.CYAN,
        "light_black":Fore.LIGHTBLACK_EX,
        "light_blue":Fore.LIGHTBLUE_EX,
        "light_cyan":Fore.LIGHTCYAN_EX,
        "light_green":Fore.LIGHTGREEN_EX,
        "light_magenta":Fore.LIGHTMAGENTA_EX,
        "light_red":Fore.LIGHTRED_EX,
        "light_white":Fore.LIGHTWHITE_EX,
        "light_yellow":Fore.LIGHTYELLOW_EX,
        "magenta":Fore.MAGENTA,
        "reset":Fore.RESET,
        "white":Fore.WHITE
    }

    back = {
        "green":Back.GREEN,
        "black":Back.BLACK,
        "red":Back.RED,
        "yellow":Back.YELLOW,
        "blue":Back.BLUE,
        "cyan":Back.CYAN,
        "light_black":Back.LIGHTBLACK_EX,
        "light_blue":Back.LIGHTBLUE_EX,
        "light_cyan":Back.LIGHTCYAN_EX,
        "light_green":Back.LIGHTGREEN_EX,
        "light_magenta":Back.LIGHTMAGENTA_EX,
        "light_red":Back.LIGHTRED_EX,
        "light_white":Back.LIGHTWHITE_EX,
        "light_yellow":Back.LIGHTYELLOW_EX,
        "magenta":Back.MAGENTA,
        "reset":Back.RESET,
        "white":Back.WHITE
    }

    You Can Always Access The Colors Like That :
        pysha.fore[...]
        pysha.back[...]
"""


def colorize(string: str) -> str:
    data = re.findall("\(Back|\(Fore\.[A-Z]*\)\[[^\[\]]*\]", string)
    print(data)
    for i in data:
        tp = i.split('.')[0][1:].strip()
        colr = i.split('.')[1].split(')')[0].strip()
        text = '['.join(i.split('[')[1:])[:-1]
        if tp.lower() == "fore":
            c = getattr(Fore, colr)
            string = string.replace(i, c + text + Fore.RESET)
        elif tp.lower() == "back":
            c = getattr(Back, colr)
            string = string.replace(i, c + text + Back.RESET)
        else:
            raise TypeError("Type Must Be Back Or Fore.")
    return string


"""
Tries To Get Possible Situations Of For Loops ( Max 3 Variable ).
"""


def make_possibles(data):
    if len(data.keys()) == 1:
        return list(range(list(data.values())[0]))
    elif len(data.keys()) == 2:
        temp = []
        for i in range(list(data.values())[0]):
            for j in range(list(data.values())[1]):
                temp.append([i, j])
        return temp
    elif len(data.keys()) == 3:
        temp = []
        for i in range(list(data.values())[0]):
            for j in range(list(data.values())[1]):
                for z in range(list(data.values())[2]):
                    temp.append([i, j, z])
        return temp
    else:
        pass


"""
Related To Loop Command. It Will Print String And Replace _i_ , _i++_ ... With The Value.
    Args :
        temp    ->     Carries Variables.
        string  ->     The String Pattern That Should Be Printed In Output.
        c       ->     If Its True, Answer Will colorized. And Else Won't Be Colorized.
"""


def display(temp: dict, string: str, c=True) -> None:
    ans = string
    for i, j in temp.items():
        ans = (ans.replace(f"_{i}_", str(j)).replace(
            f"_{i}++_", str(j+1)).replace(f"_{i}--_", str(j-1)))
    if not c:
        print(ans)
    else:
        print(colorize(ans))


"""
Related To Loop Command. It Will Get Input And Replace _i_ , _i++_ ... With The Value.
    Args :
        temp    ->     Carries Variables.
        string  ->     The String Pattern That Should Be Printed In Output.
        c       ->     If Its True, Answer Will colorized. And Else Won't Be Colorized.
"""


def get_ans(temp: dict, string: str, c=True) -> str:
    ans = string
    for i, j in temp.items():
        ans = (ans.replace(f"_{i}_", str(j)).replace(
            f"_{i}++_", str(j+1)).replace(f"_{i}--_", str(j-1)))
    if not c:
        return input(ans)
    else:
        return input(colorize(ans))


"""
All Descriptions From This Function Came In `classes.py` - command Class.
"""


def loop(cmd: str, mode='p', c=True) -> Union[None, str]:
    if mode == "p":  # Print Mode
        alls = cmd.split(">")[0][1:]
        string = cmd.split("{")[1][:-1]
        arr_data = []
        dic_tar = {}
        for i in alls.split(','):
            itemp = i.split(':')
            arr_data.append(itemp[0].strip())
            dic_tar[itemp[0].strip()] = int(itemp[1].strip())
        if len(dic_tar.keys()) >= 4:
            raise LoopError(
                "Loops With 4 Or More Variables Are Not Supported In This Version.")
        tars = make_possibles(dic_tar)
        for i in tars:
            temp = {}
            if not (type(i) == int) and len(i) == 2:
                temp = {arr_data[0]: i[0], arr_data[1]: i[1]}
            elif not (type(i) == int) and len(dic_tar) == 3:
                temp = {arr_data[0]: i[0], arr_data[1]                        : i[1], arr_data[2]: i[2]}
            else:
                temp = {arr_data[0]: i[0]}
            display(temp, string, c)
    elif mode == 'i':  # input Mode
        alls = cmd.split(">")[0][1:]
        string = cmd.split("{")[1][:-1]
        arr_data = []
        dic_tar = {}
        for i in alls.split(','):
            itemp = i.split(':')
            arr_data.append(itemp[0].strip())
            dic_tar[itemp[0].strip()] = int(itemp[1].strip())
        if len(dic_tar.keys()) >= 4:
            raise LoopError(
                "Loops With 4 Or More Variables Are Not Supported In This Version.")
        tars = make_possibles(dic_tar)
        x = 0
        y = 0
        z = 0
        ls = []
        if len(dic_tar) == 2:
            x = (list(dic_tar.values())[0])
            y = (list(dic_tar.values())[1])
            ls = [["" for _ in range(y)] for __ in range(x)]
        elif len(dic_tar) == 1:
            x = (list(dic_tar.values())[0])
            ls = ["" for _ in range(x)]
        elif len(dic_tar) == 3:
            x = (list(dic_tar.values())[0])
            y = (list(dic_tar.values())[1])
            z = (list(dic_tar.values())[2])
            ls = [[["" for _ in range(z)] for __ in range(y)]
                  for ___ in range(x)]
        x = 0
        y = 0
        z = 0
        for i in tars:
            temp = {}
            if not (type(i) == int) and len(i) == 2:
                temp = {arr_data[0]: i[0], arr_data[1]: i[1]}
                try:
                    ls[x][y] = get_ans(temp, string, c)
                    ls[x][y + 1]
                    y += 1
                except:
                    try:
                        ls[x + 1][y]
                        x += 1
                        y = 0
                    except:
                        pass
            elif not (type(i) == int) and len(i) == 3:
                temp = {arr_data[0]: i[0], arr_data[1]                        : i[1], arr_data[2]: i[2]}
                try:
                    ls[x][y][z] = get_ans(temp, string, c)
                    ls[x][y][z + 1]
                    z += 1
                except:
                    try:
                        ls[x][y + 1][z]
                        y += 1
                        z = 0
                    except:
                        try:
                            ls[x + 1][y][z]
                            x += 1
                            y = 0
                            z = 0
                        except:
                            pass
            else:
                temp = {arr_data[0]: i}
                try:
                    ls[x] = get_ans(temp, string, c)
                    ls[x]
                    x += 1
                except:
                    pass
        return ls
    else:
        raise ModeError("Mode Should Be 'p' Or 'i' !")


"""
Descrioptions Are Available In 'classes.py' And In command Class.
"""


def condition(check: str, p=True, **kwargs) -> Union[None, str, int]:
    if not re.match("""(\"?)[ -~]+(\"?)(>|<|==|>=|<=)(\"?)[ -~]+(\"?)\?(\"?)[ -~]+(\"?):(\"?)[ -~]+(\"?)""", ''.join(check.split(' '))):
        raise ConditionError("An Error In Parsing Condition :(")
    if "eval" in check:
        if not check[check.find("eval") - 1:].startswith("\eval"):
            raise SecurityError("Using This Funcion Is Forbidden.")
    if "exec" in check:
        if not check[check.find("exec") - 1:].startswith("\exec"):
            raise SecurityError("Using This Funcion Is Forbidden.")
    index_of_ok_sign = not_between_index("?", check)
    p1 = index_split(check, index_of_ok_sign)[0].strip()
    p2 = index_split(check, index_of_ok_sign)[1].strip()
    ans = eval(p1, kwargs, {})
    index_of_ok_sign = not_between_index(":", p2)
    ok = index_split(p2, index_of_ok_sign)[0].strip()
    not_ok = index_split(p2, index_of_ok_sign)[1].strip()
    if ans:
        if p:
            print(eval(ok, kwargs))
        else:
            return eval(ok, kwargs)
    else:
        if p:
            print(eval(not_ok, kwargs))
        else:
            return eval(not_ok, kwargs)


"""
    You Can Write Into A File :)
    Args :
        file_name
        text
    Returns True If Its Successfully Done, Else Return False
"""


def write_file(file_name: str, text: Union[str, list]) -> bool:
    if type(text) == str:
        try:
            a = open(file_name, "w")
            a.write(text)
            a.close()
            return True
        except:
            return False
    elif type(text) == list:
        try:
            a = open(file_name, "w")
            a.writelines(text)
            a.close()
            return True
        except:
            return False


"""
    You Can Read Data From A File :)
    Args :
        file_name

        mode(Default:'s') :   The Mode Can Be 's' Or 'l' Means List Or String.
                            If You Want To Get Output As String Mode 's' Will
                            Be Good For You And If You Want To Get As list
                            Mode 'l' Is Good For You.

    Returns list of lines Of The File If mode Is 'l', string of lines
    If mode is 's'.
"""


def read_file(file_name: str, mode="s") -> Union[list, str]:
    try:
        a = open(file_name, "r")
        d = a.readlines()
        a.close()
        if mode == "s":
            return "".join(d)
        elif mode == "l":
            return d
        else:
            raise ModeError(
                f"Mode {mode} Not Found ! Mode Should Be 's'(string) Or 'l'(List)")
    except:
        return ""


"""
    You Can Append to A File.
    Args :
        file_name
        text
    Returns True If Its Successfully Done, Else Return False
"""


def append_file(file_name: str, text: str) -> bool:
    if type(text) == str:
        try:
            a = open(file_name, "a")
            a.write(text)
            a.close()
            return True
        except:
            return False
    elif type(text) == list:
        try:
            a = open(file_name, "a")
            a.writelines(text)
            a.close()
            return True
        except:
            return False


"""
    Tries To Create A Directory.
    Args :
        dir_name ->  Directory Name. You Can Pass An String Like 'test' Or 
                   'test/test1/test2' For Directory. But If You Want To 
                   Make Parents Too, You Have To Pass Next Argument Manually
                   Too.

        create_parents(Default:False) ->  If Its False, Parents Will Not Be
                                            Created If They Are Not Exists. 
                                            And If Its True, Parents Will Be
                                            Created Too !

    Returns True If Its Successfully Done, Else Return False
"""


def create_dir(dir_name: str, create_parents=False) -> bool:
    if create_parents:
        try:
            makedirs(dir_name)
            return True
        except:
            return False
    else:
        try:
            mkdir(dir_name)
            return True
        except:
            return False


"""
    Tries To Replace A Dictionary.
    Args :
        file_name ->  File Name. You Can Pass File Name That You Want Data 
                     To Change.

        data_to_replace ->  Data That You Want To Replace As Dictinary.
                            For Example : {'test':'test1'} Will Replace All
                            'test' With 'test1' In The File.

    Returns True If Its Successfully Done, Else Return False
"""


def replace_in_file(file_name: str, data_to_replace: dict) -> bool:
    f = open(file_name, "r")
    data = f.readlines()
    f.close()
    for i in data:
        for key, value in data_to_replace.items():
            if key in i:
                i = i.replace(key, value)
    f = open(file_name, "w")
    f.writelines(data)
    f.close()


"""
    Tries To Remove A Directory.
    Args :
        dir_name ->  Directory Name. You Can Pass An String Like 'test' Or 
                   'test/test1/test2' For Directory. If You Want To Force
                   Remove That Childs Will Remove Too, Pass Next Argument
                   Manually.

        force(Default:False) ->  If Its False And If directory Has Childs,
                                Directory Won't Be Removed.But If Its False
                                Doesn't Matter What We Are Cold-Blood Murderes
                                And We Kill Them All :)

    Returns True If Its Successfully Done, Else Return False
"""


def rm_dir(dir_name: str, force=False) -> bool:
    if not force:
        try:
            rmdir(dir_name)
            return True
        except:
            return False
    else:
        try:
            rmtree(dir_name)
            return True
        except:
            return False


"""
    Checks If An String Exists Between Two Characters In Another String.
    Args :
        string ->   A Simple String. (We Need To Check If This String Exists
                    In Another String Or Not) For Example If You Want To Check
                    If 'hello' is between " Characters, You Need To Pass hello
                    As This Argument.

        string_to_check ->   A Big String That We Want To Check And Dive In :)
                            For Example 'hello My Name Is Arshia And "hello" Is
                            ok Now As i think :)' And Pass It As This Argument.
        
        start_sign -> This Should Be The Start Character That You Are Looking For.
                    For Example You Can Pass " As This Argument ( Continue Of Previous
                    Part Example ) . But In Another Example That You Want To Check A Tag For
                    Example html Tag You Can Pass "<" As Start.
        
        end_sign -> This Should Be The Start Character That You Are Looking For.
                    For Example You Can Pass " As This Argument ( Continue Of Previous
                    Part Example ) . Continue Of Tag Example You Can Pass ">" As End.
        
        exact -> Basically, Checks For <html> Not <*html*>.This Argument Will Check
                For exact Value. For Example If This Option Was False And You Want
                To Check For Tags In  String "This <html Test Fake Text> Is Ok." It
                Will Return True But If This Option Is True,For That String Will
                Return False. Because It Looks For Exact "<html>" Not < Then Anything
                Then html Then Again Anything And > Character.
        
        case_sensetive -> If Its True, Lower Case Or Upper Case In Words Are Important.
                         For Example If Its True, "Html" Is Different From "html" But
                         If Its False, "Html" And "html" Are Equal.
        

    Example :
        1. between(":",'The : Is Not Between "" Chars.','"')
        1. Returns :
            1. False
        2. between(":",'The ":" Is Between "" Chars !','"')
        2. Returns :
            2. True

    Returns True If Its Found In target Text Else, Returns False.
"""


def between(string, string_to_check, start_sign='"', end_sign='"', exact=False, case_sensetive=True):
    is_in = False
    ts = ""
    for c, i in enumerate(string_to_check):
        if is_in:
            if i == end_sign:
                is_in = False
                ts = ""
        else:
            if i == start_sign:
                is_in = True
        if is_in:
            ts += i
        if not exact:
            if case_sensetive:
                if string in ts:
                    return True
            else:
                if string.lower() in ts.lower():
                    return True
        else:
            if case_sensetive:
                if start_sign + string + end_sign in ts:
                    return True
            else:
                if (start_sign + string + end_sign).lower() in ts.lower():
                    return True
    return False


"""
    Acts Like Between Function But It Will Return Index Insted Of Boolean Value.
"""


def between_index(string, string_to_check, start_sign='"', end_sign='"', exact=False, case_sensetive=True, start=0):
    is_in = False
    ts = ""
    for c, i in enumerate(string_to_check[start:]):
        if is_in:
            if i == end_sign:
                is_in = False
                ts = ""
        else:
            if i == start_sign:
                is_in = True
        if is_in:
            ts += i
        if not exact:
            if case_sensetive:
                if string in ts:
                    return (c - len(string) + 1 + start)
            else:
                if string.lower() in ts.lower():
                    return (c - len(string) + 1 + start)
        else:
            if case_sensetive:
                if start_sign + string + end_sign in ts:
                    return (c - len(string) + 1 + start)
            else:
                if (start_sign + string + end_sign).lower() in ts.lower():
                    return (c - len(string) + 1 + start)
    return -1


"""
    Reverse Of Between Plus Index Of Target.
    Args :
        string ->   A Simple String. (We Need To Check If This String Exists
                    In Another String Or Not) For Example If You Want To Check
                    If 'hello' is not between " Characters, You Need To Pass hello
                    As This Argument.

        string_to_check ->   A Big String That We Want To Check And Dive In :)
                            For Example 'hello My Name Is Arshia And "hello" Is
                            ok Now As i think :)' And Pass It As This Argument.
                
        start_sign -> This Should Be The Start Character That You Are Looking For.
                    For Example You Can Pass " As This Argument ( Continue Of Previous
                    Part Example ) . But In Another Example That You Want To Check Text
                    That Does Not In A Tag For Example html Tag You Can Pass "<" As Start.
        
        end_sign -> This Should Be The Start Character That You Are Looking For.
                    For Example You Can Pass " As This Argument ( Continue Of Previous
                    Part Example ) . Continue Of Tag Example You Can Pass ">" As End.
        
        case_sensetive -> If Its True, Lower Case Or Upper Case In Words Are Important.
                         For Example If Its True, "Html" Is Different From "html" But
                         If Its False, "Html" And "html" Are Equal.
    
    Example :
        1. not_between_index(":",'Arshia Said : "Hello ! There Is : Hiding Here :)"','"','"')
        1. Returns :
            1. 12
        2. not_between_index(":",'":" Is Not Out Of "" Sign. But Now : It Is.','"','"')
        2. Returns :
            2. 35

    Returns True If Its Found Out Of target Sign, Else Returns False.
"""


def not_between_index(string, string_to_check, start_sign='"', end_sign='"', case_sensetive=True, start=0):
    is_in = False
    ts = ""
    for c, i in enumerate(string_to_check[start:]):
        if is_in:
            if i == end_sign:
                is_in = False
        else:
            if i == start_sign:
                is_in = True
                ts = ""
        if not is_in:
            ts += i
        if case_sensetive:
            if string in ts:
                return (c - len(string) + 1 + start)
        else:
            if string.lower() in ts.lower():
                return (c - len(string) + 1 + start)
    return -1


"""
    Split String By Index.
    Example :
        index_split("Hello My Name Is Arshia",10)
    Returns :
        ['Hello My N', 'me Is Arshia']
"""


def index_split(string, ind):
    return [string[:ind], string[ind + 1:]]


"""
    This Function Will Try To Loop A Function In Specific count/infinite count.
    Example :
        delay_loop(10,func,5,args,kwargs)
    
    In Every 10 Seconds Function func(args,kwargs) Will Be Called 5 Times.
"""


def delay_loop(delay: int, func: Callable, count: int = -1, *args, **kwargs):
    if count == -1:
        index = 0
        while True:
            time.sleep(delay)
            index += 1
            func(index, *args, **kwargs)
    else:
        for index in range(count):
            time.sleep(delay)
            func(index, *args, **kwargs)


"""
    This Function Will Merge Some List Or Some Dict Or Some Dict And List !
    Example :
        merge(["a","b"],{"c":"d","e":"f"},["g","h"])
    Returns :
        {"a":"","b":"","c":"","d":"","e":"","f":"","g":"","h":""}

    Example :
        merge(["a","b"],{"c":"d","e":"f"},["g","h"],force_list=True)
    Returns :
        ["a","b","c","e","d","f","g","h"]

    Example :
        merge(["a","b"],{"c":"d","e":"f"},["g","h"],force_list=True,dont_change=True)
    Returns :
        ["a","b","c","d","e","f","g","h"]
    
    Example:
        merge(["a","b"],["c","d","e","f"],["g","h"])
    Returns :
        ["a","b","c","d","e","f","g","h"]
    
    In Every 10 Seconds Function func(args,kwargs) Will Be Called 5 Times.
"""


def merge(*args, force_list=False, dont_change=False):
    if list(filter(lambda x: type(x) == dict, args)):
        if not force_list:
            ans = {}
            for i in args:
                if type(i) == list:
                    ans[i] = ""
                elif type(i) == dict:
                    ans.update(i)
            return ans
        else:
            ans = []
            for i in args:
                if type(i) == list:
                    ans.extend(i)
                elif type(i) == dict:
                    if dont_change:
                        for a, b in i.items():
                            ans.extend([a, b])
                    else:
                        ans.extend(i.keys())
                        ans.extend(i.values())
            return ans
    else:
        ans = []
        for i in args:
            ans.extend(i)
        return ans



from .functions import *
import getpass

# Authentication Class
#  Mode 0 -> Not Show Password
#  Mode 1 -> Star Password
#  Mode 2 -> Normal Input


class Auth:
    def __init__(self, first_prompt="Enter Your Username : ", second_prompt="Enter Your Password : ", mode=0, mask='*'):
        self.first_prompt = first_prompt
        self.second_prompt = second_prompt
        self.mode = mode
        self.mask = mask

    """
    Argumets : 
        c           ->  If Its True colorize Function Will Be Executed On Prompts.
        mask_color  ->  Set Color For Password Mask.
        inp_color   ->  Set Color For Username Mask.
        mask        ->  Set Mask Character.
    """

    def auth(self, c=True, mask_color=fore["reset"], inp_color=fore["reset"], mask='*'):
        tar = self.mask
        if mask != '*':
            tar = mask
        if len(tar.strip()):
            pass
        if c:
            ans = []
            if self.mode == 0:
                uname = colorprompt(
                    colorize(self.first_prompt), char_color=inp_color)
                pwd = getpass.getpass(colorize(self.second_prompt))
                ans.append(uname)
                ans.append(pwd)
            elif self.mode == 1:
                uname = colorprompt(
                    colorize(self.first_prompt), char_color=inp_color)
                pwd = passprompt(colorize(self.second_prompt),
                                 mask_color=mask_color, mask=mask)
                ans.append(uname)
                ans.append(pwd)
            elif self.mode == 2:
                uname = colorprompt(
                    colorize(self.first_prompt), char_color=inp_color)
                pwd = input(prompt=colorize(self.second_prompt))
                ans.append(uname)
                ans.append(pwd)
            else:
                raise ModeError(
                    f"Mode {self.mode} Is Not Available For Authentication. Just 0,1,2.")
            return ans
        else:
            ans = []
            if self.mode == 0:
                uname = colorprompt((self.first_prompt), char_color=inp_color)
                pwd = getpass.getpass((self.second_prompt))
                ans.append(uname)
                ans.append(pwd)
            elif self.mode == 1:
                uname = colorprompt((self.first_prompt), char_color=inp_color)
                pwd = passprompt(prompt=(self.second_prompt),
                                 mask=mask_color + tar + fore["reset"])
                ans.append(uname)
                ans.append(pwd)
            elif self.mode == 2:
                uname = colorprompt((self.first_prompt), char_color=inp_color)
                pwd = colorprompt((self.second_prompt), char_color=inp_color)
                ans.append(uname)
                ans.append(pwd)
            else:
                raise ModeError(
                    f"Mode {self.mode} Is Not Available For Authentication. Just 0,1,2.")
            return ans



import sys
import datetime

from .data_structures import PyshaString


"""
This Function Acts As A Decorator For Retring On Some Exceptions That You Define.
For Example You Can Set That If Running A Function Have HTTPConnectionError, Function
Retries And Run Again. You Can Define Count Of Retries Too.

Agrs : 
    count : Count Of Retries For Target Function
    exceptions : a list contains all exceptions that should retry on them or '*' character
                means all of exceptions should be retried.

"""


def retry(count, exceptions='*'):
    def decorator(func):
        def wrapper(*args, **kwargs):
            tested = 0
            while tested < count:
                try:
                    func(*args, **kwargs)
                except:
                    if exceptions == '*':
                        tested += 1
                        continue
                    inf = sys.exc_info()
                    if inf[0] in exceptions:
                        tested += 1
                        continue
                    else:
                        raise
        return wrapper
    return decorator


"""
This Function Acts As A Decorator For Ignoring Some Exceptions That You Define.
For Example You Can Set That If Running A Function Have ZeroDivisionError, Function
Ignore The Problem And Not Make Exception.

Agrs : 
    exceptions That Should Be A List Of Exceptions You Want To Ignore Or '*' Character
     That Means All Of Exceptions Must Ignore.
"""


def ignore(exceptions='*'):
    def decorator(func):
        def wrapper(*args, **kwargs):
            try:
                func(*args, **kwargs)
            except:
                if exceptions == '*':
                    return
                inf = sys.exc_info()
                if inf[0] in exceptions:
                    return
                else:
                    raise
        return wrapper
    return decorator


"""
This Function Acts As A Decorator For Process Initialization. Every Function
That Has This Decorator Will Be Called Before The Main Function Once. Important
Thing Is That Function Must Not Have Any Arguments Or Kwarguments To Be Called.

Important Thing : 
    * This Function Will Be Called Before The Main Function.
    * Function Must Not Have Any Arguments Or Kwarguments To Be Called.
"""


def once(func):
    func()

    def decorator(*args, **kwargs):
        func(*args, **kwargs)
    return decorator


"""
This Function Acts As A Decorator For Check Spent Time On Executing Functions.
Args:
    Gets One Argument For Printing Format. For Example :
        @time("Spent Time Was _T_")
        def test():
            return True
    And Output Of This Code After Calling test() Will Be Something Like That : 
        Spent Time Was 0:00:00.100404

Important Parameters : 
    _T_ : Will Be Replaced With [Hour]:[Minute]:[Second].[MiliSecond]
    _H_ : Will Be Replaced With [Hour]
    _M_ : Will Be Replaced With [Minute]
    _S_ : Will Be Replaced With [Second]
    _MS_ : Will Be Replaced With [MiliSecond]

Important :
    You Can Escape All Of This Parameters With (\)

"""


def timer(string):
    def sec(func):
        def run(*args, **kwargs):
            start = datetime.datetime.now()
            func(*args, **kwargs)
            end = datetime.datetime.now()
            ans = PyshaString(string)
            s = ans.replace_with_escape("_T_", str(end - start))
            s = s.replace_with_escape("_S_", str(end - start).split(':')
                                      [2].split('.')[0].strip())
            try:
                s = s.replace_with_escape("_MS_", str(end - start).split(':')
                                          [2].split('.')[1].strip())
            except:
                s = s.replace_with_escape("_MS_", "0")
            s = s.replace_with_escape("_M_", str(
                end - start).split(':')[1].strip())
            s = s.replace_with_escape("_H_", str(
                end - start).split(':')[0].strip())
            print(s)
        return run
    return sec



from colorama import Fore,Back

fore = {
    "green" : Fore.GREEN,
    "black":Fore.BLACK,
    "red":Fore.RED,
    "yellow":Fore.YELLOW,
    "blue":Fore.BLUE,
    "cyan":Fore.CYAN,
    "light_black":Fore.LIGHTBLACK_EX,
    "light_blue":Fore.LIGHTBLUE_EX,
    "light_cyan":Fore.LIGHTCYAN_EX,
    "light_green":Fore.LIGHTGREEN_EX,
    "light_magenta":Fore.LIGHTMAGENTA_EX,
    "light_red":Fore.LIGHTRED_EX,
    "light_white":Fore.LIGHTWHITE_EX,
    "light_yellow":Fore.LIGHTYELLOW_EX,
    "magenta":Fore.MAGENTA,
    "reset":Fore.RESET,
    "white":Fore.WHITE
}

back = {
    "green":Back.GREEN,
    "black":Back.BLACK,
    "red":Back.RED,
    "yellow":Back.YELLOW,
    "blue":Back.BLUE,
    "cyan":Back.CYAN,
    "light_black":Back.LIGHTBLACK_EX,
    "light_blue":Back.LIGHTBLUE_EX,
    "light_cyan":Back.LIGHTCYAN_EX,
    "light_green":Back.LIGHTGREEN_EX,
    "light_magenta":Back.LIGHTMAGENTA_EX,
    "light_red":Back.LIGHTRED_EX,
    "light_white":Back.LIGHTWHITE_EX,
    "light_yellow":Back.LIGHTYELLOW_EX,
    "magenta":Back.MAGENTA,
    "reset":Back.RESET,
    "white":Back.WHITE
}